<?php

declare(strict_types=1);

namespace App\Controller\OgAgent;

use App\Controller\OgRepository\Git\CreateRepositoryAction;
use App\Controller\OgRepository\Git\CreateTagAction;
use App\Entity\Client;
use App\Entity\Command;
use App\Entity\GitImageRepository;
use App\Entity\Image;
use App\Entity\ImageImageRepository;
use App\Entity\ImageRepository;
use App\Entity\Partition;
use App\Entity\Trace;
use App\Model\ClientStatus;
use App\Model\CommandTypes;
use App\Model\ImageStatus;
use App\Model\PartitionTypes;
use App\Model\TraceStatus;
use App\Service\Trace\CreateService;
use Doctrine\ORM\EntityManagerInterface;
use Exception;
use Psr\Log\LoggerInterface;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Attribute\Route;
use Symfony\Component\Validator\Exception\ValidatorException;
use Symfony\Contracts\HttpClient\Exception\ClientExceptionInterface;
use Symfony\Contracts\HttpClient\Exception\RedirectionExceptionInterface;
use Symfony\Contracts\HttpClient\Exception\ServerExceptionInterface;
use Symfony\Contracts\HttpClient\Exception\TransportExceptionInterface;
use Symfony\Contracts\HttpClient\HttpClientInterface;

class CreateImageAction extends AbstractOgAgentController
{
    /**
     * @throws TransportExceptionInterface
     * @throws ServerExceptionInterface
     * @throws RedirectionExceptionInterface
     * @throws ClientExceptionInterface
     */
    public function __invoke(Image $image, ?Partition $partition = null, ?Client $client = null): JsonResponse
    {
        if (!$image->getClient()->getIp()) {
            throw new ValidatorException('IP is required');
        }

        $partitionInfo = [];

        if ($partition) {
            $partitionInfo["numPartition"] = $partition->getPartitionNumber();
            $partitionInfo["numDisk"] = $partition->getDiskNumber();
            $partitionInfo["partitionCode"] = $partition->getPartitionCode();
            $partitionInfo["filesystem"] = $partition->getFilesystem();
            $partitionInfo["osName"] = $partition->getOperativeSystem()?->getName();
            $image->setPartitionInfo(json_encode($partitionInfo));
        } else {
            $partitionInfo = json_decode($image->getPartitionInfo(), true);
        }

        if ($image->getType() === 'monolithic') {
            $repository = $image->getClient()->getRepository();
            $latestImageRepo = $this->entityManager->getRepository(ImageImageRepository::class)->findLatestVersionByImageAndRepository($image, $repository);

            $imageImageRepository = new ImageImageRepository();
            $imageImageRepository->setName($image->getName().'_v'.($latestImageRepo ? $latestImageRepo->getVersion() + 1 : 1));
            $imageImageRepository->setImage($image);
            $imageImageRepository->setRepository($repository);
            $imageImageRepository->setStatus(ImageStatus::IN_PROGRESS);
            $imageImageRepository->setVersion($latestImageRepo ? $latestImageRepo->getVersion() + 1 : 1);
            $imageImageRepository->setPartitionInfo(json_encode($partitionInfo));

            $this->entityManager->persist($imageImageRepository);

            return $this->createMonolithicImage($imageImageRepository, $partitionInfo, $image, $repository, $client);
        } else {
            $repository = $image->getClient()->getRepository();
            $latestImageRepo = $this->entityManager->getRepository(GitImageRepository::class)->findLatestVersionByImageAndRepository($image, $repository);

            $gitImageRepository = new GitImageRepository();
            $gitImageRepository->setName($image->getName().'_v'.($latestImageRepo ? $latestImageRepo->getVersion() + 1 : 1));
            $gitImageRepository->setImage($image);
            $gitImageRepository->setRepository($repository);
            $gitImageRepository->setStatus(ImageStatus::IN_PROGRESS);
            $gitImageRepository->setVersion($latestImageRepo ? $latestImageRepo->getVersion() + 1 : 1);
            $gitImageRepository->setTag('v'.($latestImageRepo ? $latestImageRepo->getVersion() + 1 : 1));
            $gitImageRepository->setCreated(false);

            $this->entityManager->persist($gitImageRepository);
            $this->entityManager->persist($image);
            $this->entityManager->flush();

            return $this->createGitImage($gitImageRepository, $partitionInfo, $image, $repository);
        }
    }

    /**
     * @throws TransportExceptionInterface
     * @throws ServerExceptionInterface
     * @throws RedirectionExceptionInterface
     * @throws ClientExceptionInterface
     */
    public function createMonolithicImage(
        ImageImageRepository $imageImageRepository,
        array $partitionInfo,
        Image $image,
        ImageRepository $repository,
        ?Client $client = null
    ): JsonResponse
    {
        $data = [
            'dsk' => (string) $partitionInfo['numDisk'],
            'par' => (string) $partitionInfo['numPartition'],
            'cpt' => null,
            'idi' => $imageImageRepository->getUuid(),
            'nci' => $image->getName().'_v'.$imageImageRepository->getVersion(),
            'ipr' => $repository->getIp(),
            'nfn' => 'CrearImagen',
            'ids' => '0'
        ];

        $partitionTypes = PartitionTypes::getPartitionTypes();

        $partitionCode = $partitionInfo['partitionCode'];
        $cptKey = array_search($partitionCode, array_column($partitionTypes, 'name'), true);

        if ($cptKey !== false) {
            $keys = array_keys($partitionTypes);
            $partitionTypeCode = $keys[$cptKey];

            $data['cpt'] = dechex($partitionTypeCode);
        } else {
            throw new Exception("El tipo de partición '$partitionCode' no se encontró en la lista.");
        }

        $client = $client ?? $image->getClient();
        $this->logger->info('Creating image', ['image' => $image->getId()]);

        $response = $this->createRequest(
            method: 'POST',
            url: 'https://'.$client->getIp().':8000/opengnsys/CrearImagen',
            params: [
                'json' => $data,
            ],
            token: $client->getToken(),
        );

        $this->logger->info('Creating image', ['image' => $imageImageRepository->getName(), 'repository' => $repository->getIp()]);

        if (isset($response['error']) && $response['code'] === Response::HTTP_INTERNAL_SERVER_ERROR) {
            throw new ValidatorException('Error creating image');
        }

        $jobId = $response['job_id'];

        $client->setStatus(ClientStatus::BUSY);
        $this->entityManager->persist($client);
        $this->entityManager->flush();

        $inputData = [
            'method' => 'CrearImagen',
            'type' => 'monolithic',
            'client' => $client->getUuid(),
            'image' => $image->getUuid(),
            'partitionCode' => $partitionInfo['partitionCode'],
            'partitionType' => $partitionInfo['filesystem'],
            'repository' => $repository->getIp(),
            'name' => $image->getName().'_v'.$imageImageRepository->getVersion(),
        ];

        $this->createService->__invoke($image->getClient(), CommandTypes::CREATE_IMAGE, TraceStatus::IN_PROGRESS, $jobId, $inputData);

        return new JsonResponse(data: $image, status: Response::HTTP_OK);
    }

    public function createGitImage(
        GitImageRepository $gitImageRepository,
        array $partitionInfo,
        Image $image,
        ImageRepository $repository
    ): JsonResponse
    {
        if (!isset($partitonInfo)) {
            try {
                $this->createRepositoryAction->__invoke($image, $repository);
            } catch (Exception $e) {
                $this->logger->error('Error creating repository', ['repository' => $repository->getId(), 'error' => $e->getMessage()]);

                return new JsonResponse(
                    data: ['error' => $e->getMessage()],
                    status: Response::HTTP_INTERNAL_SERVER_ERROR
                );
            }
        } else {
            try {
                $this->createTagAction->__invoke($image, $repository, $gitImageRepository);
            } catch (Exception $e) {
                $this->logger->error('Error creating tag', ['repository' => $repository->getId(), 'error' => $e->getMessage()]);

                return new JsonResponse(
                    data: ['error' => $e->getMessage()],
                    status: Response::HTTP_INTERNAL_SERVER_ERROR
                );
            }
        }

        return new JsonResponse(
            data: ['message' => 'Repository created successfully'],
            status: Response::HTTP_OK
        );

        //TODO: llamar al endpoint del agente.
    }
}
